dofile("./lua/mmfi.lua")
dofile("./lua/win.lua")

Conf = require "./lua/conf"
Profile	= require ("./lua/profiles/" .. Conf.defaultProfile)

local util		= require "./lua/util"
local frame		= require "./lua/frame"

Fileio = require "./lua/fileio"

Frame = frame.FrameClass()

editor	= require "./lua/editor"

Fileio.editor = editor

function EditorInit ()
	Frame:Init()
	editor.InitLayers()
	
	EditorReInit()
end

function EditorReInit ()
	editor.controls.toolbar:Init()
	editor.controls.layerToolbar:Init()
	editor.controls.grid:Init()
	--editor.controls.tilePalette:Init()
	
	for k,v in pairs(editor.layers) do
		v:Init()
	end
	
	editor.vars.layer:Activate()
end

function EditorReset ()
	Frame:Init ()
	for k,v in pairs(editor.layers) do
		v:Reset()
	end
	
	editor.layers = { }
end

function EditorUpdate ()
	editor.frame:Scroll()
	
	-- Update control positions and sizes
	editor.controls.toolbar:Update()
	editor.controls.statusbar:Update()
	editor.controls.sidebar:Update()
	editor.controls.layerToolbar:Update()
	editor.controls.clientArea:Update()
	editor.controls.clientVScroll:Update()
	editor.controls.clientHScroll:Update()
	editor.controls.sidebarVScroll:Update()
	editor.controls.sidebarHScroll:Update()
	editor.controls.clientScrollCorner:Update()
	editor.controls.divider:Update()
	editor.controls.clientCanvas:Update()
	editor.controls.grid:Update()
	editor.controls.viewport:Update()
	editor.controls.statusInfo:Update()
	
	editor.controls.selectBox:Update()
	editor.controls.floatingBox:Update()
	
	editor.vars.layer:UpdateControls()
end

function EditorStep ()
	editor.cursor:SetPosition(mmf.Mouse.GetX(), mmf.Mouse.GetY())
	
	editor.cursor:Step()
	--editor.layerCursor:Step()
	
	if (Frame.updateQueued == true) then
		Frame:UpdateProperties()
		editor.SetModified()
	end
	
	editor.controls.divider:Step()
	editor.controls.selectBox:Step()
	editor.controls.floatingBox:Step()
	
	editor.vars.layer:StepControls()
	editor.vars.layer:Step()
	
	if (editor.controls.divider.engagable or editor.controls.divider.engaged) then
		editor.cursor:SetState('vdiv')
	elseif (editor.controls.clientArea:ClientUnderMouse(editor.cursor)) then
		--editor.cursor:SetState(editor.vars.layer.cursor)
	else
		editor.cursor:SetState('win')
	end
	
	editor.cursor:PanScroll()
	
end

-- ----------------------------------------------------------------------------
-- MMF-instatntiated Interrupts

Interrupt = { }

function Interrupt.OnScroll (fixed, scrollpos)
	if (editor.fixedTable[fixed] == nil) then
		return
	end
	return editor.fixedTable[fixed]:SetScrollPosition(scrollpos)
	--editor.fixedTable[fixed]:SetScrollPosition(scrollpos)
end

function Interrupt.ButtonPressed (fixed)
	if (editor.fixedTable[fixed] == nil) then
		return
	end
	return editor.fixedTable[fixed]:OnButtonPress()
end

function FileDropped (filename)
	Fileio.OpenGivenFile(filename)
end

-- ----------------------------------------------------------------------------
-- MMF Menu Calls

Menu = { }

function Menu.FileNew ()
	if (Fileio.PromptSaveFile() == false) then
		return
	end
	EditorReset()
	EditorInit()
end

function Menu.FileOpen ()
	Fileio:Open()
end

function Menu.FileSave ()
	Fileio:Save()
end

function Menu.FileSaveAs ()
	Fileio:SaveAs()
end

function Menu.FileExit ()
	if (Fileio.PromptSaveFile() == false) then
		return
	end
	MF_Exit()
end

function Menu.EditUndo ()
	editor.vars.layer:Undo()
end

function Menu.EditRedo ()
	editor.vars.layer:Redo()
end

function Menu.EditCut ()
	editor.vars.layer:Cut()
end

function Menu.EditCopy ()
	editor.vars.layer:Copy()
end

function Menu.EditPaste ()
	editor.vars.layer:Paste()
end

function Menu.EditFrameProperties ()
	Frame:OpenProperties()
end

function Menu.ViewZoomIn ()
	editor.controls.toolbar.tbZoomIn:OnButtonPress()
end

function Menu.ViewZoomOut ()
	editor.controls.toolbar.tbZoomOut:OnButtonPress()
end

function Menu.ViewGrid ()
	editor.controls.grid:Toggle()
end

function Menu.LayerProperties ()
	editor.vars.layer:OpenProperties()
end
