local setmetatable = setmetatable
local getmetatable = getmetatable
local error = error
local tostring = tostring

module (...)

-- ----------------------------------------------------------------------------

Deque = {
	first = 0,
	last = -1,
}

function Deque:New ()
	obj = { }
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Deque:PushFront (value)
	self.first = self.first - 1
	self[self.first] = value
end

function Deque:PushBack (value)
	self.last = self.last + 1
	self[self.last] = value
end

function Deque:PopFront ()
	if (self.first > self.last) then
		error ("Deque.PopFront called on empty deque", 2)
	end
	
	local value = self[self.first]
	self[self.first] = nil
	self.first = self.first + 1
	return value
end

function Deque:PopBack ()
	if (self.first > self.last) then
		error ("Deque.PopBack called on empty deque", 2)
	end
	
	local value = self[self.last]
	self[self.last] = nil
	self.last = self.last - 1
	return value
end

function Deque:Front ()
	if (self.first > self.last) then
		error ("Deque.Front called on empty deque", 2)
	end
	return self[self.front]
end

function Deque:Back ()
	if (self.first > self.last) then
		error("Deque.Back called on empty deque", 2)
	end
	return self[self.last]
end

function Deque:Size ()
	return self.last - self.first + 1
end

function Deque:Empty ()
	if (self:Size() > 0) then
		return false
	end
	
	return true
end

function Deque:Clear ()
	self = Deque:New()
end

-- ----------------------------------------------------------------------------

SlidingDeque = Deque:New({
	deque = nil,
	max_size = 0,
})

function SlidingDeque:New (size)
	obj = { max_size = size, deque = Deque:New() }
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function SlidingDeque:SetSize (sz)
	self.max_size = sz
	while (self.deque:Size() > sz) do
		self.deque:PopFront()
	end
end

function SlidingDeque:PushFront (value)
	if (self.deque:Size() >= self.max_size) then
		self.deque:PopBack()
	end
	self.deque:PushFront(value)
end

function SlidingDeque:PushBack (value)
	if(self.deque:Size() >= self.max_size) then
		self.deque:PopFront()
	end
	self.deque:PushBack(value)
end

function SlidingDeque:PopFront ()
	return self.deque:PopFront()
end

function SlidingDeque:PopBack ()
	return self.deque:PopBack()
end

function SlidingDeque:Front ()
	return self.deque:Front()
end

function SlidingDeque:Back ()
	return self.deque:Back()
end

function SlidingDeque:Size ()
	return self.deque:Size()
end

function SlidingDeque:Empty ()
	return self.deque:Empty()
end

function SlidingDeque:Clear ()
	self.deque = Deque:New()
end

-- ----------------------------------------------------------------------------

Queue = { }

function Queue:New ()
	obj = { deque = Deque:New() }
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Queue:Push (value)
	self.deque:PushBack(value)
end

function Queue:Pop ()
	if (self.deque:Empty()) then
		error("Queue.Pop called on empty queue", 2)
	end
	return self.deque:PopFront()
end

function Queue:Front ()
	if (self.deque:Empty()) then
		error("Queue.Front called on empty queue", 2)
	end
	return self.deque:Front()
end

function Queue:Size ()
	return self.deque:Size()
end

function Queue:Empty ()
	return self.deque:Empty()
end

function Queue:Clear ()
	self.deque:Clear()
end

-- ----------------------------------------------------------------------------

Stack = { }

function Stack:New ()
	obj = { deque = Deque:New() }
	setmetatable (obj, self)
	self.__index = self
	return obj
end

function Stack:Push (value)
	self.deque:PushBack(value)
end

function Stack:Pop ()
	if (self.deque:Empty()) then
		error("Stack.Pop called on empty stack", 2)
	end
	return self.deque:PopBack()
end

function Stack:Top ()
	if (self.deque:Empty()) then
		error("Stack.Top called on empty stack", 2)
	end
	return self.deque:Back()
end

function Stack:Size ()
	return self.deque:Size()
end

function Stack:Empty ()
	return self.deque:Empty()
end

function Stack:Clear ()
	return self.deque:Clear()
end