local util = require "./lua/util"

local setmetatable = setmetatable
local require = require
local print = print
local pairs = pairs
local error = error
local math = math
local table = table

local mmf = mmf
local win = win
local Profile = Profile
local Fileio = Fileio

local Frame = Frame

local MF_Error = MF_Error

module (...)

-- ----------------------------------------------------------------------------

local CLASS = require "./lua/editor_com"
local LAYER = require "./lua/editor_layer"

-- Conf
conf = { 
	baseLayer = 1,
}

-- Struct Declarations
layers = { }
controls = { }
frame = { }

-- Editor Variables
vars = { 
	filepath = "",
	filename = "Untitled.tlv",
	iofilter = "",
	modified = false,
}

--LAYER.current = layers[4]
--LAYER.SetCurrent(layers[4])

-- Cursor Object
cursor = CLASS.CursorObject()
cursor.stateTable['move'] = 0
cursor.stateTable['draw'] = 12
cursor.stateTable['erase'] = 13
cursor.stateTable['select'] = 14
cursor.stateTable['pointer'] = 15
cursor.stateTable['vdiv'] = 16
cursor.stateTable['stamp'] = 17
cursor.stateTable['fill'] = 18
cursor.stateTable['pan'] = 19
cursor.LDown = false
cursor.MDown = false
cursor.RDown = false
cursor.canvas = CLASS.CursorObject( {xtile = 0, ytile = 0} )
cursor.layer = CLASS.CursorObject()

--layerCursor = CLASS.CursorObject( {xtile = 0, ytile = 0} )

-- Share common structs
LAYER.controls = controls
LAYER.cursor = cursor
--LAYER.layerCursor = layerCursor

function SetModified ()
	if (vars.modified == false) then
		vars.modified = true
		mmf.Window.SetTitle("Tile Lattice - " .. vars.filename .. "*")
	end
end

function ClearModified ()
	if (vars.modified == true) then
		vars.modified = false
		mmf.Window.SetTitle("Tile Lattice - " .. vars.filename)
	end
end

function TileWidth ()
	if (vars.layer == nil) then
		return Frame.tileWidth
	end
	local tileProps = vars.layer:TileProperties()
	return tileProps.tileWidth or Frame.tileWidth
end

function TileHeight ()
	if (vars.layer == nil) then
		return Frame.tileHeight
	end
	local tileProps = vars.layer:TileProperties()
	return tileProps.tileHeight or Frame.tileHeight
end

function cursor:Step ()
	if (mmf.Mouse.ButtonDown(1)) then
		self.LDown = true
	else
		self.LDown = false
	end
	
	if (mmf.Mouse.ButtonDown(2)) then
		self.MDown = true
	else
		self.MDown = false
	end
	
	if (mmf.Mouse.ButtonDown(3)) then
		self.RDown = true
	else
		self.RDown = false
	end
	
	self.canvas.LDown = self.LDown
	self.canvas.MDown = self.MDown
	self.canvas.RDown = self.RDown
	
	self.layer.LDown = self.LDown
	self.layer.MDown = self.MDown
	self.layer.RDown = self.RDown
	
	-- Layer coordinate translation
	self.layer.x = (self.x - controls.clientArea.clientx) / controls.viewport.zoom + controls.clientHScroll.position
	self.layer.y = (self.y - controls.clientArea.clienty) / controls.viewport.zoom + controls.clientVScroll.position
	
	-- TODO
	-- Canvas coordinate translation
	self.canvas.x = self.layer.x - controls.clientArea.reserved --+ controls.clientArea.reservedEX --math.min(0, controls.clientArea.xoverflow) / 2
	self.canvas.y = self.layer.y - controls.clientArea.reserved --+ controls.clientArea.reservedEY --math.min(0, controls.clientArea.yoverflow) / 2
	self.canvas.xtile = math.floor(self.canvas.x / TileWidth())
	self.canvas.ytile = math.floor(self.canvas.y / TileHeight())
end

function cursor:PanScroll ()
	if (self.MDown == false) then
		self.panLatch = false
		return
	elseif (controls.clientArea:ClientUnderMouse(self) and self.MDown and self.panLatch == false) then
		self.panX = self.x
		self.panY = self.y
		self.panLatch = true
	elseif (self.panLatch == false) then
		return
	end

	local xdist = self.x - self.panX
	local ydist = self.y - self.panY
	
	controls.clientHScroll:SetScrollPosition(controls.clientHScroll.position + math.min(16, xdist / 8))
	controls.clientVScroll:SetScrollPosition(controls.clientVScroll.position + math.min(16, ydist / 8))
	
	self:SetState('pan')
end

-- Frame Object

frame.xleft = 0
frame.xright = 0
frame.ytop = 0
frame.ybottom = 0

function frame:Scroll ()
	--self.xleft = controls.clientHScroll.position
	self.xright = self.xleft + mmf.Window.GetCWidth()
	--self.ytop = controls.clientVScroll.position
	self.ybottom = self.ytop + mmf.Window.GetCHeight()
	
	local xoff = mmf.Window.GetCWidth() / 2
	local yoff = mmf.Window.GetCHeight() / 2

	--mmf.Frame.CenterAt(xoff + self.xleft, yoff + self.ytop)
end

-- Control Objects
controls.toolbar = CLASS.ControlObject( {h = 32} )
controls.statusbar = CLASS.ControlObject( {h = 20} )
controls.sidebar = CLASS.ControlObject( {w = 180, clientx = 0, clienty = 0, xCurs = 0, yCurs = 0} )
controls.clientArea = CLASS.ControlObject( {clientx = 0, clienty = 0, reserved = 160, reservedEX = 0, reservedEY = 0, framew = 0, frameh = 0, xoverflow = 0, yoverflow = 0} )
controls.clientVScroll = CLASS.ScrollbarObject( {w = 16} )
controls.clientHScroll = CLASS.ScrollbarObject( {h = 16} )
controls.sidebarVScroll = CLASS.ScrollbarObject( {w = 16} )
controls.sidebarHScroll = CLASS.ScrollbarObject( {h = 16} )
controls.clientScrollCorner = CLASS.ControlObject( )
controls.divider = CLASS.ControlObject( {w = 4} )
controls.clientCanvas = CLASS.ControlObject( {color = 0, xCurs = 0, yCurs = 0, xTileCurs = 0, yTileCurs = 0} )
controls.grid = CLASS.GridObject( {mode = 1} )
controls.viewport = CLASS.ViewportObject( {zoom = 1} )
controls.statusInfo = CLASS.ControlObject( )
controls.tilePaletteSource = CLASS.ActivePictureObject( )
controls.tilePalette = CLASS.ClippedObject( {tileWidth = 0, tileHeight = 0, xOrigin = 0, yOrigin = 0, xSpacing = 1, ySpacing = 1} )
controls.sidebarSelect = CLASS.ControlObject( )
controls.layerToolbar = CLASS.ControlObject( {w = 32} )
controls.selectBox = CLASS.ControlObject( {ox = 0, oy = 0, sx = 0, sy = 0, dw = 0, dh = 0, active = false, snapping = false} )
controls.floatingBox = CLASS.ControlObject( {cx = 0, cy = 0, px = 0, py = 0, sx = 0, sy = 0, dw = 0, dh = 0, active = false, snapping = false, held = false} )

-- Toolbar Controls
controls.toolbar.tbNew = CLASS.ControlObject( {h = 28, w = 28} )
controls.toolbar.tbOpen = CLASS.ControlObject( {h = 28, w = 28} )
controls.toolbar.tbSave = CLASS.ControlObject( {h = 28, w = 28} )
controls.toolbar.tbDiv1 = CLASS.ControlObject( {h = 28, w = 2} )
controls.toolbar.tbUndo = CLASS.ControlObject( {h = 28, w = 28} )
controls.toolbar.tbRedo = CLASS.ControlObject( {h = 28, w = 28} )
controls.toolbar.tbDiv2 = CLASS.ControlObject( {h = 28, w = 2} )
controls.toolbar.tbCut = CLASS.ControlObject( {h = 28, w = 28} )
controls.toolbar.tbCopy = CLASS.ControlObject( {h = 28, w = 28} )
controls.toolbar.tbPaste = CLASS.ControlObject( {h = 28, w = 28} )
controls.toolbar.tbDiv3 = CLASS.ControlObject( {h = 28, w = 2} )
controls.toolbar.tbZoomIn = CLASS.ControlObject( {h = 28, w = 28} )
controls.toolbar.tbZoomOut = CLASS.ControlObject( {h = 28, w = 28} )
controls.toolbar.tbDiv4 = CLASS.ControlObject( {h = 28, w = 2} )

-- Layer Toolbar Controls

controls.layerToolbar.tbDivTop = CLASS.ControlObject( {w = 32, h = 2} )
controls.layerToolbar.tbDivBottom = CLASS.ControlObject( {w = 32, h = 2} )

function InitLayers (layerset)
	local layerlist = { }
	if (layerset ~= nil) then
		layerlist = util.paramsort(layerset, 'order')
	else
		layerlist = util.paramsort(Profile.layers, 'order')
	end
	
	for i = 1, #layerlist do
		if (layerlist[i].type == 'tile') then
			layers[i] = LAYER.TileLayerObject(layerlist[i])
			layers[i].layerType = 1
		elseif (layerlist[i].type == 'object') then
			layers[i] = LAYER.ObjectLayerObject(layerlist[i])
			layers[i].layerType = 2
		elseif (layerlist[i].type == 'backdrop') then
			layers[i] = LAYER.BackdropLayerObject(layerlist[i])
			layers[i].layerType = 3
		else
			layers[i] = LAYER.LayerObject(layerlist[i])
		end

		layers[i].layerName = layerlist[i].name
	end

	vars.layer = layers[Profile.startLayer]
	LAYER.SetCurrent(vars.layer)
end

-- Control Init Subroutines

function controls.toolbar:Init ()
	self.tbUndo:Disable()
	self.tbRedo:Disable()
	self.tbCut:Disable()
	self.tbCopy:Disable()
	self.tbPaste:Disable()
end

function controls.layerToolbar:Init ()
	local buttoninfo = { }
	
	-- Generate layer list and button list
	for i = 1, #layers do
		local bid = 'layer' .. i

		if (layers[i].layerType == 1) then
			if (layers[i].subtype == 'collision') then
				buttoninfo[bid] = {i, layers[i].layerName, mmf.Create.LTBTileObsticle}
			elseif (layers[i].subtype == 'platform') then
				buttoninfo[bid] = {i, layers[i].layerName, mmf.Create.LTBTilePlatform}
			elseif (layers[i].subtype == 'normal' or layers[i].subtype == nil) then
				buttoninfo[bid] = {i, layers[i].layerName, mmf.Create.LTBTile}
			end

		elseif (layers[i].layerType == 2) then
			buttoninfo[bid] = {i, layers[i].layerName, mmf.Create.LTBObject}
		elseif (layers[i].layerType == 3) then
			buttoninfo[bid] = {i, layers[i].layerName, mmf.Create.LTBBackdrop}
		end
	end

	-- If old buttons exist, destroy them
	if (self.buttons ~= nil) then
		for k,v in pairs (self.buttons) do
			mmf.Object.Destroy(v.expID)
		end
	end

	self.buttons = { }
	for k,v in pairs(buttoninfo) do
		local layerButton = CLASS.ControlObject( {h = 28, w = 28, order = v[1]} )
		self.buttons[k] = layerButton
		
		local expID = table.maxn(exportTable) + 1
		exportTable[expID] = self.buttons[k]
		self.buttons[k].expID = expID
		
		v[3](expID)
		mmf.Control.SetTooltip(self.buttons[k].fixedID, v[2])
		
		if (vars.layer.order == v[1]) then
			self.buttons[k]:Check()
		end
		
		-- Button control
		function layerButton:OnButtonPress ()
			vars.layer:Deactivate()
			
			vars.layer = layers[self.order]
			LAYER.SetCurrent(vars.layer)
			
			for k,v in pairs(controls.layerToolbar.buttons) do
				v:Uncheck()
			end
			self:Check()
			
			vars.layer:Activate()
		end
	end
	
	
end

function controls.grid:Init ()
	self:SetColor(0xCCCCCC)
	self:SetCellSize(TileWidth(), TileHeight())
	self:SetMode(0)
end

-- Control Update Subroutines

function controls.toolbar:Update ()
	self.x = frame.xleft -- mmf.Frame.GetXLeft()
	self.y = frame.ytop -- mmf.Frame.GetYTop()
	self.w = mmf.Window.GetCWidth()
	
	-- Button List
	local buttons = {
		'tbNew', 'tbOpen', 'tbSave', 'tbDiv1',
		'tbUndo', 'tbRedo', 'tbDiv2',
		'tbCut', 'tbCopy', 'tbPaste', 'tbDiv3',
		'tbZoomOut', 'tbZoomIn', 'tbDiv4',
	}
	
	-- Update button positions
	self.tbNew.x = self.x
	self.tbOpen.x = self.tbNew.x + self.tbNew.w
	self.tbSave.x = self.tbOpen.x + self.tbOpen.w
	self.tbDiv1.x = self.tbSave.x + self.tbSave.w + 2
	self.tbUndo.x = self.tbDiv1.x + self.tbDiv1.w + 2
	self.tbRedo.x = self.tbUndo.x + self.tbUndo.w
	self.tbDiv2.x = self.tbRedo.x + self.tbRedo.w + 2
	self.tbCut.x = self.tbDiv2.x + self.tbDiv2.w + 2
	self.tbCopy.x = self.tbCut.x + self.tbCut.w
	self.tbPaste.x = self.tbCopy.x + self.tbCopy.w
	self.tbDiv3.x = self.tbPaste.x + self.tbPaste.w + 2
	self.tbZoomOut.x = self.tbDiv3.x + self.tbDiv3.w + 2
	self.tbZoomIn.x = self.tbZoomOut.x + self.tbZoomOut.w
	self.tbDiv4.x = self.tbZoomIn.x + self.tbZoomIn.w + 2
	
	vars.layer:UpdateTools(self.tbDiv4.x + self.tbDiv4.w + 2, self.y + 2)

	for k,v in pairs(buttons) do
		self[v].y = self.y + 2
		self[v]:SetPosition()
	end
	
	self:SetPosition()
	self:SetSize()
end

function controls.statusbar:Update ()
	self.x = frame.xleft -- mmf.Frame.GetXLeft()
	self.y = frame.ybottom - self.h -- mmf.Frame.GetYBottom() - self.h
	self.w = mmf.Window.GetCWidth()
	
	self:SetPosition()
	self:SetSize()
end

function controls.sidebar:Update ()
	self.x = frame.xleft -- mmf.Frame.GetXLeft()
	self.y = frame.ytop + controls.toolbar.h -- mmf.Frame.GetYTop() + controls.toolbar.h
	self.h = mmf.Window.GetCHeight() - controls.toolbar.h - controls.statusbar.h
	
	self.clientx = self.x + 1
	self.clienty = self.y + 1
	self.clientw = self.w - controls.sidebarVScroll.w - 2
	self.clienth = self.h - controls.sidebarHScroll.h - 2
	
	self.xCurs = mmf.Mouse.GetX() - self.clientx + controls.sidebarHScroll.position
	self.yCurs = mmf.Mouse.GetY() - self.clienty + controls.sidebarVScroll.position
	
	self:SetPosition()
	self:SetSize()
end

function controls.layerToolbar:Update ()
	self.x = frame.xright - self.w
	self.y = frame.ytop + controls.toolbar.h
	self.h = mmf.Window.GetCHeight() - controls.toolbar.h - controls.statusbar.h
	
	self.tbDivTop.x = self.x
	self.tbDivBottom.x = self.x
	
	self.tbDivTop.y = self.y
	self.tbDivBottom.y = self.y + self.h - 2
	
	self.tbDivTop:Update()
	self.tbDivBottom:Update()
	
	-- Buttons

	local buttonlist = util.paramsort(self.buttons, 'order')
	for i = 1, #buttonlist do
		buttonlist[i].x = self.x + 3
		buttonlist[i].y = self.y + 4 + (#buttonlist - i) * 28
		buttonlist[i]:Update()
	end
	
	self:SetPosition()
	self:SetSize()
end

function controls.clientArea:Update ()
	self.x = controls.sidebar.x + controls.sidebar.w + controls.divider.w
	self.y = controls.sidebar.y
	self.w = mmf.Window.GetCWidth() - controls.sidebar.w - controls.divider.w - controls.layerToolbar.w
	self.h = controls.sidebar.h
	
	self.clientx = self.x + 1
	self.clienty = self.y + 1
	self.clientw = self.w - controls.clientVScroll.w - 2
	self.clienth = self.h - controls.clientHScroll.h - 2
	
	-- TODO
	-- Set some parameters about the usable frame area
	self.framew = self.reserved * 2 + Frame.frameWidth
	self.frameh = self.reserved * 2 + Frame.frameHeight
	self.xoverflow = self.framew - (self.clientw / controls.viewport.zoom)
	self.yoverflow = self.frameh - (self.clienth / controls.viewport.zoom)
	
	-- If window is too big, we pad the client area (and redraw everything, oh no!)
	self.reservedEX = 0
	self.reservedEY = 0
	if (self.xoverflow < 0) then
		self.reservedEX = 0 - (self.xoverflow / 2)
	end
	if (self.yoverflow < 0) then
		self.reservedEY = 0 - (self.yoverflow / 2)
	end
	
	self:SetPosition()
	self:SetSize()
end

function controls.clientVScroll:Update ()
	self.x = controls.clientArea.x + controls.clientArea.w - self.w - 1
	self.y = controls.clientArea.y + 1
	self.h = controls.clientArea.h - controls.clientHScroll.h - 2
	
	self:SetPosition()
	self:SetSize()
end

function controls.clientHScroll:Update ()
	self.x = controls.clientArea.x + 1
	self.y = controls.clientArea.y + controls.clientArea.h - self.h - 1
	self.w = controls.clientArea.w - controls.clientVScroll.w - 2
	
	self:SetPosition()
	self:SetSize()
end

function controls.sidebarVScroll:Update ()
	self.x = controls.sidebar.x + controls.sidebar.w - self.w - 1
	self.y = controls.sidebar.y + 1
	self.h = controls.sidebar.h - controls.sidebarHScroll.h - 2
	
	self:SetPosition()
	self:SetSize()
end

function controls.sidebarHScroll:Update ()
	self.x = controls.sidebar.x + 1
	self.y = controls.sidebar.y + controls.sidebar.h - self.h - 1
	self.w = controls.sidebar.w - controls.sidebarVScroll.w - 2
	
	self:SetPosition()
	self:SetSize()
end

function controls.clientScrollCorner:Update ()
	self.x = controls.clientVScroll.x
	self.y = controls.clientHScroll.y
	self.w = controls.clientVScroll.w
	self.h = controls.clientHScroll.h
	
	self:SetPosition()
	self:SetSize()
end

function controls.divider:Update ()
	self.x = controls.sidebar.x + controls.sidebar.w
	self.y = controls.sidebarVScroll.y - 1
	self.h = controls.sidebarVScroll.h + controls.sidebarHScroll.h + 2
	
	self:SetPosition()
	self:SetSize()
end

function controls.clientCanvas:Update ()
	-- If window is too big, we pad the client area (and redraw everything, oh no!)
	--[[if (controls.clientArea.xoverflow < 0) then
		controls.clientArea.reservedEX = 0 - (controls.clientArea.xoverflow / 2)
		--self.x = self.x - controls.clientArea.xoverflow / 2
	end
	if (controls.clientArea.yoverflow < 0) then
		controls.clientArea.reservedEY = 0 - (controls.clientArea.yoverflow / 2)
		--self.y = self.y - controls.clientArea.yoverflow / 2
	end]]
	
	self.x = controls.clientArea.reserved --+ controls.clientArea.reservedEX
	self.y = controls.clientArea.reserved --+ controls.clientArea.reservedEY
	self.w = Frame.frameWidth
	self.h = Frame.frameHeight

	-- If window is too big, we pad the client area (and redraw everything, oh no!)
	--[[if (controls.clientArea.xoverflow < 0) then
		controls.clientArea.reservedEX = 0 - (controls.clientArea.xoverflow / 2)
		self.x = self.x - controls.clientArea.xoverflow / 2
	end
	if (controls.clientArea.yoverflow < 0) then
		controls.clientArea.reservedEY = 0 - (controls.clientArea.yoverflow / 2)
		self.y = self.y - controls.clientArea.yoverflow / 2
	end]]
	
	-- Cursor location information
	self.xCurs = mmf.Mouse.GetX() - self.x - controls.clientArea.clientx + controls.clientHScroll.position
	self.yCurs = mmf.Mouse.GetY() - self.y - controls.clientArea.clienty + controls.clientVScroll.position
	self.xTileCurs = math.floor(self.xCurs / TileWidth())
	self.yTileCurs = math.floor(self.yCurs / TileHeight())
	
	if (self.color ~= Frame.backgroundColor) then
		self.color = Frame.backgroundColor
		self:SetColor(self.color)
	end

	self:SetPosition()
	self:SetSize()
end

function controls.grid:Update ()
	self.x = controls.clientArea.clientx
	self.y = controls.clientArea.clienty
	self.w = controls.clientArea.clientw + self.wCell
	self.h = controls.clientArea.clienth + self.hCell
	
	self.wCell = TileWidth() * controls.viewport.zoom
	self.hCell = TileHeight() * controls.viewport.zoom
	
	local diffx = (TileWidth() - ((controls.clientCanvas.x - controls.clientHScroll.position) % TileWidth())) * controls.viewport.zoom
	local diffy = (TileHeight() - ((controls.clientCanvas.y - controls.clientVScroll.position) % TileHeight())) * controls.viewport.zoom

	self:SetPosition(self.x - diffx, self.y - diffy)
	self:SetSize()
	self:SetCellSize()
end

function controls.viewport:Update ()
	self.x = controls.clientArea.clientx
	self.y = controls.clientArea.clienty
	self.w = controls.clientArea.clientw
	self.h = controls.clientArea.clienth

	self.xSource = controls.clientHScroll.position
	self.ySource = controls.clientVScroll.position
	self.wSource = self.w / self.zoom
	self.hSource = self.h / self.zoom
	
	local adjusted_width = controls.clientArea.framew
	local adjusted_height = controls.clientArea.frameh
	
	-- Scrollbars probably need updating
	if (controls.clientArea.xoverflow <= 0) then
		adjusted_width = adjusted_width - controls.clientArea.xoverflow
		controls.clientHScroll:Disable()
		controls.clientHScroll:SetScrollPosition(0)
	else
		controls.clientHScroll:Enable()
		controls.clientHScroll:SetScrollRange(0, math.abs(controls.clientArea.xoverflow))
	end
	
	if (controls.clientArea.yoverflow <= 0) then
		adjusted_height = adjusted_height - controls.clientArea.yoverflow
		controls.clientVScroll:Disable()
		controls.clientVScroll:SetScrollPosition(0)
	else
		controls.clientVScroll:Enable()
		controls.clientVScroll:SetScrollRange(0, math.abs(controls.clientArea.yoverflow))
	end
	
	-- We also need to update the window size
	local real_width = math.max(adjusted_width, mmf.Window.GetCWidth())
	local real_height = math.max(adjusted_height, mmf.Window.GetCHeight())
	
	if (real_width ~= mmf.Frame.GetWidth()) then
		mmf.Frame.SetWidth(real_width)
	end
	if (real_width ~= mmf.Window.GetFrameWidth()) then
		mmf.Window.SetFrameWidth(real_width)
	end
	if (real_height ~= mmf.Frame.GetHeight()) then
		mmf.Frame.SetHeight(real_width)
	end
	if (real_height ~= mmf.Window.GetFrameHeight()) then
		mmf.Window.SetFrameHeight(real_height)
	end

	self:SetPosition()
	self:SetSize()
	self:SetSource()
end

function controls.statusInfo:Update ()
	self.x = controls.statusbar.x
	self.y = controls.statusbar.y + 2
	self.w = controls.statusbar.w
	self.h = controls.statusbar.h - 2
	
	if (controls.clientArea:ClientUnderMouse(cursor)) then
		controls.statusInfo:SetText('(' .. cursor.canvas.xtile .. ', ' .. cursor.canvas.ytile .. ')')
	elseif (controls.sidebar:ClientUnderMouse(cursor)) then
		controls.statusInfo:SetText('Source (' .. controls.sidebar.xCurs .. ', ' .. controls.sidebar.yCurs .. ')')
	else
		controls.statusInfo:SetText('')
	end
	
	self:SetPosition()
	self:SetSize()
end

-- Control Step Subroutines

function controls.divider:Step ()
	self:TestEngage(cursor)
	
	-- If the control is engaged, we're dragging the divider to resize the sidebar
	if (self.engaged) then
		local w = mmf.Frame.GetXLeft() + mmf.Mouse.GetClientX() - 2
		if (w < (controls.sidebarHScroll.h * 2 + controls.sidebarVScroll.w + 2)) then
			w = controls.sidebarHScroll.h * 2 + controls.sidebarVScroll.w + 2
		end
		if (w > (mmf.Frame.GetXRight() - controls.clientHScroll.h * 2 - controls.clientVScroll.w - 6)) then
			w = mmf.Frame.GetXRight() - controls.clientHScroll.h * 2 - controls.clientVScroll.w - 6
		end
		controls.sidebar:SetSize(w)
	end
end

-- Toolbar Buttons Pressed

function controls.toolbar.tbZoomIn:OnButtonPress ()
	if (controls.viewport.zoom == .5) then
		controls.viewport.zoom = 1
		controls.toolbar.tbZoomOut:Enable()
		mmf.Menu.Enable("ViewZoomOut")
	elseif (controls.viewport.zoom == 1) then
		controls.viewport.zoom = 2
	elseif (controls.viewport.zoom == 2) then
		controls.viewport.zoom = 4
		self:Disable()
		mmf.Menu.Disable("ViewZoomIn")
	end
end

function controls.toolbar.tbZoomOut:OnButtonPress ()
	if (controls.viewport.zoom == 4) then
		controls.viewport.zoom = 2
		controls.toolbar.tbZoomIn:Enable()
		mmf.Menu.Enable("ViewZoomIn")
	elseif (controls.viewport.zoom == 2) then
		controls.viewport.zoom = 1
	elseif (controls.viewport.zoom == 1) then
		controls.viewport.zoom = .5
		self:Disable()
		mmf.Menu.Disable("ViewZoomOut")
	end
end

function controls.toolbar.tbNew:OnButtonPress ()
	if (Fileio.PromptSaveFile() == false) then
		return
	end
	Frame:Init()
	
	for k,v in pairs(layers) do
		v:Reset()
	end
	layers = { }
	
	InitLayers()
	controls.toolbar:Init()
	controls.layerToolbar:Init()
	controls.grid:Init()
	
	for k,v in pairs(layers) do
		v:Init()
	end
	vars.layer:Activate()
end

function controls.toolbar.tbOpen:OnButtonPress ()
	Fileio:Open()
end

function controls.toolbar.tbSave:OnButtonPress ()
	Fileio:Save()
end

function controls.toolbar.tbCut:OnButtonPress ()
	vars.layer:Cut()
end

function controls.toolbar.tbCopy:OnButtonPress ()
	vars.layer:Copy()
end

function controls.toolbar.tbPaste:OnButtonPress ()
	vars.layer:Paste()
end

function controls.toolbar.tbUndo:OnButtonPress ()
	vars.layer:Undo()
end

function controls.toolbar.tbRedo:OnButtonPress ()
	vars.layer:Redo()
end

-- Menu Options

function controls.grid:Toggle ()
	if (controls.grid.visible) then
		controls.grid:Hide()
		mmf.Menu.Uncheck("ViewGrid")
	else
		controls.grid:Show()
		mmf.Menu.Check("ViewGrid")
	end
end

-- Select Box stuff - TODO: Probably move somewhere else

function controls.selectBox:StartSnapping (x, y)
	if (self.active) then
		return
	end
	
	self.snapping = true
	self:Start(x, y)
end

function controls.selectBox:Start (x, y)
	if (self.active) then
		return
	end
	
	self.active = true
	self.ox = x
	self.oy = y
	self.sx = x
	self.sy = y
end

function controls.selectBox:Stop ()
	self.active = false
	self.snapping = false
	self.ox = 0
	self.oy = 0
	self.sx = -1
	self.sy = -1
	self.dw = 0
	self.dh = 0
end

function controls.selectBox:Step ()
	if (self.active ~= true) then
		return
	end
	
	if (self.snapping) then
		self.dw = cursor.canvas.xtile - self.ox
		self.dh = cursor.canvas.ytile - self.oy
		if (cursor.canvas.x - cursor.canvas.xtile * TileWidth() > TileWidth() / 2) then
			self.dw = self.dw + 1
		end
		if (cursor.canvas.y - cursor.canvas.ytile * TileHeight() > TileHeight() / 2) then
			self.dh = self.dh + 1
		end
	else
		self.dw = cursor.canvas.x - self.ox
		self.dh = cursor.canvas.y - self.oy
	end
	
	AutoScroll(48)
end

function controls.selectBox:Update ()
	if (self.active ~= true) then
		mmf.Object.SetVisibility(self.exportID, false)
		return
	end
	
	-- TODO
	if (self.snapping) then
		self.x = self.sx * TileWidth() + controls.clientArea.reserved -- controls.clientArea.reservedEX
		self.y = self.sy * TileHeight() + controls.clientArea.reserved -- controls.clientArea.reservedEY
		self.w = self.dw * TileWidth()
		self.h = self.dh * TileHeight()
	else
		self.x = self.sx + controls.clientArea.reserved -- controls.clientArea.reservedEX
		self.y = self.sy + controls.clientArea.reserved -- controls.clientArea.reservedEY
		self.w = self.dw
		self.h = self.dh
	end
	
	if (self.w < 0) then
		self.x = self.x + self.w
		self.w = 0 - self.w
	end
	if (self.h < 0) then
		self.y = self.y + self.h
		self.h = 0 - self.h
	end
	
	self:SetPosition()
	self:SetSize()
	
	mmf.Object.SetVisibility(self.exportID, true)
end

function controls.selectBox:Coords ()
	local coords = { 
		sx = self.sx,
		sy = self.sy,
		dw = self.dw,
		dh = self.dh
	}
	
	if (coords.dw < 0) then
		coords.sx = coords.sx + coords.dw
		coords.dw = 0 - coords.dw
	end
	if (coords.dh < 0) then
		coords.sy = coords.sy + coords.dh
		coords.dh = 0 - coords.dh
	end
	
	return coords
end

function controls.floatingBox:StartSnapping (x, y, w, h)
	if (self.active) then
		return
	end
	
	self.snapping = true
	self:Start(x, y, w, h)
end

function controls.floatingBox:Start (x, y, w, h)
	if (self.active) then
		return
	end
	
	self.active = true
	self.sx = x
	self.sy = y
	self.dw = w
	self.dh = h
end

function controls.floatingBox:Stop ()
	self.active = false
	self.snapping = false
	self.autoscroll = false
	self.type = nil
end

function controls.floatingBox:Grab (x, y)
	if (self.held) then
		return
	end
	
	if (x == nil and self.snapping) then
		x = cursor.canvas.xtile
	elseif (x == nil) then
		x = cursor.canvas.x
	end
	
	if (y == nil and self.snapping) then
		y = cursor.canvas.ytile
	elseif (y == nil) then
		y = cursor.canvas.y
	end
	
	self.cx = x - self.sx
	self.cy = y - self.sy
	
	self.held = true
end

function controls.floatingBox:Release ()
	self.held = false
end

function controls.floatingBox:Step ()
	if (self.active == false or self.held == false) then
		return
	end
	
	self.px = self.sx
	self.py = self.sy

	if (self.snapping) then
		self.sx = cursor.canvas.xtile - self.cx
		self.sy = cursor.canvas.ytile - self.cy
	else
		self.sx = cursor.canvas.x - self.cx
		self.sy = cursor.canvas.y - self.cy
	end
	
	if (self.autoscroll) then
		AutoScroll(64)
	end
end

function controls.floatingBox:Update ()
	if (self.active ~= true) then
		mmf.Object.SetVisibility(self.exportID, false)
		return
	end
	
	-- TODO
	if (self.snapping) then
		self.x = self.sx * TileWidth() + controls.clientArea.reserved --+ controls.clientArea.reservedEX
		self.y = self.sy * TileHeight() + controls.clientArea.reserved --+ controls.clientArea.reservedEY
		self.w = self.dw * TileWidth()
		self.h = self.dh * TileHeight()
	else
		self.x = self.sx + controls.clientArea.reserved --+ controls.clientArea.reservedEX
		self.y = self.sy + controls.clientArea.reserved --+ controls.clientArea.reservedEY
		self.w = self.dw
		self.h = self.dh
	end
	
	self:SetPosition()
	self:SetSize()
	
	mmf.Object.SetVisibility(self.exportID, true)
end

-- Independent functions

function AutoScroll (range)
	if (cursor.x > controls.clientArea.clientx + controls.clientArea.clientw - range) then
		local xdist = cursor.x - (controls.clientArea.clientx + controls.clientArea.clientw - range)
		controls.clientHScroll:SetScrollPosition(controls.clientHScroll.position + math.min(12, xdist / 16))
	elseif (cursor.x < controls.clientArea.clientx + range) then	
		local xdist = (controls.clientArea.clientx + range) - cursor.x
		controls.clientHScroll:SetScrollPosition(controls.clientHScroll.position - math.min(12, xdist / 16))
	end
	
	if (cursor.y > controls.clientArea.clienty + controls.clientArea.clienth - range) then
		local ydist = cursor.y - (controls.clientArea.clienty + controls.clientArea.clienth - range)
		controls.clientVScroll:SetScrollPosition(controls.clientVScroll.position + math.min(12, ydist / 16))
	elseif (cursor.y < controls.clientArea.clienty + range) then	
		local ydist = (controls.clientArea.clienty + range) - cursor.y
		controls.clientVScroll:SetScrollPosition(controls.clientVScroll.position - math.min(12, ydist / 16))
	end
end

-- Callback processing

exportTable = { }
fixedTable = { }

LAYER.BindConf(conf)
LAYER.BindControls(controls)
LAYER.BindCursor(cursor)
LAYER.BindFixedTable(fixedTable)

LAYER.fixedTable = fixedTable

exportTable[1] = controls.toolbar
exportTable[2] = controls.statusbar
exportTable[3] = controls.sidebar
exportTable[4] = controls.clientArea
exportTable[5] = controls.clientVScroll
exportTable[6] = controls.clientHScroll
exportTable[7] = controls.sidebarVScroll
exportTable[8] = controls.sidebarHScroll
exportTable[9] = controls.clientScrollCorner
exportTable[10] = controls.divider
exportTable[11] = controls.clientCanvas
exportTable[12] = cursor
exportTable[13] = controls.grid
exportTable[14] = controls.viewport
exportTable[15] = controls.statusInfo
exportTable[16] = controls.toolbar.tbNew
exportTable[17] = controls.toolbar.tbOpen
exportTable[18] = controls.toolbar.tbSave
exportTable[19] = controls.toolbar.tbDiv1
exportTable[20] = controls.toolbar.tbUndo
exportTable[21] = controls.toolbar.tbRedo
exportTable[22] = controls.toolbar.tbDiv2
exportTable[23] = controls.toolbar.tbCut
exportTable[24] = controls.toolbar.tbCopy
exportTable[25] = controls.toolbar.tbPaste
exportTable[26] = controls.toolbar.tbDiv3
exportTable[27] = controls.toolbar.tbZoomIn
exportTable[28] = controls.toolbar.tbZoomOut
exportTable[29] = controls.toolbar.tbDiv4
LAYER.TileLayer.Exports.tilePaletteSource = 30
LAYER.TileLayer.Exports.tilePalette = 31
LAYER.TileLayer.Exports.tileSelect = 32
exportTable[33] = controls.layerToolbar
exportTable[34] = controls.layerToolbar.tbDivTop
exportTable[35] = controls.layerToolbar.tbDivBottom
LAYER.BaseLayer.Exports.tbSelect = 36
LAYER.BaseLayer.Exports.tbDraw = 37
LAYER.BaseLayer.Exports.tbErase = 38
exportTable[39] = controls.selectBox
exportTable[40] = controls.floatingBox
LAYER.BaseLayer.Exports.tbStamp = 41
LAYER.BaseLayer.Exports.tbFill = 42
LAYER.BaseLayer.Exports.tbNewObject = 43
LAYER.BaseLayer.Exports.tbNewBackdrop = 44
exportTable[45] = 0

mmf.ExportCallback = function (obj)
	if (exportTable[obj] ~= nil) then
		exportTable[obj]:Bind(obj)
		fixedTable[exportTable[obj].fixedID] = exportTable[obj]
	end
end