local CLASS = require "./lua/editor_com"
local Container = require "./lua/container"
local Undo = require "./lua/undo"

local util = require "./lua/util"

local Base = require "./lua/editor_baselayer"

local Profile = Profile

module(..., package.seeall)

-- ----------------------------------------------------------------------------
-- Unbound objects from the editor

Controls = { }
Current = nil
Cursor = { }
FixedTable = { }

-- ----------------------------------------------------------------------------

Layer = Base.Layer:new( {
	undoStack = nil,
	undoPool = nil,
} )

function Layer:new (obj)
	obj = obj or {}
	setmetatable (obj, self)
	self.__index = self
	obj:Construct()
	return obj
end

function Layer:Construct ()
	self.layerType = 3
	self.float = { }
	self.undoStack = Undo.UndoStack:New()
	self.undoPool = nil
	
	-- Construct Controls
	
	self.tools.tbNewBackdrop = CLASS.ControlObject( {parent = self, w = 28, h = 28, name = 'newbackdrop'} )	
end

function Layer:Init ()
	self.tools.tbNewBackdrop:Bind(Base.Exports.tbNewBackdrop)
end

function Layer:Activate ()
	for k,v in pairs(self.tools) do
		v:Disable()
		v:Enable()
	end
	
	-- Sidebar adjustment
	--local width = math.min(self.controls.tilePalette.wSource, math.floor(mmf.Window.GetCWidth() / 4))
	--Controls.sidebar.w = math.max(Controls.sidebarHScroll.h * 2, width) + Controls.sidebarVScroll.w + 2
	
	-- Rebind controls
	FixedTable[self.tools.tbNewBackdrop.fixedID] = self.tools.tbNewBackdrop

	-- Why is this necessary?
	self.tools.tbNewBackdrop.parent = self
end

function Layer:Deactivate ()
	for k,v in pairs(self.tools) do
		v:SetPosition(0, -64)
		v:Disable()
		--v:Uncheck()
	end
	
	Controls.sidebarHScroll:Disable()
	Controls.sidebarVScroll:Disable()
end

function Layer:Reload ()
	for k,v in pairs(self.tools) do
		v:Disable()
	end
	self:Activate()
end

function Layer:Reset ()
	self:Deactivate()
end

function Layer:UpdateTools (x, y)
	-- Button List
	local buttons = {
		'tbNewBackdrop',
	}

	-- Update button positions
	self.tools.tbNewBackdrop.x = x

	for k,v in pairs(buttons) do
		self.tools[v].y = y
		self.tools[v]:SetPosition()
	end
end

function Layer:Step ()
	if (win.state.menusOpen > 0 or win.state.active == 0) then
		return
	end
	
	--if (self.queuedUpdate) then
	--	self:UpdateProperties()
	--end

	Controls.clientArea:TestEngage(Cursor)
	
	-- Update cursor
	if (Controls.clientArea.engagable or Controls.clientArea.engaged) then
		Cursor:SetState(self.cursor)
	else
		Cursor:SetState('win')
	end
	
	-- Undo state
	if (self.undoStack:UndoAvailable()) then
		Controls.toolbar.tbUndo:Enable()
		mmf.Menu.Enable("EditUndo")
	else
		Controls.toolbar.tbUndo:Disable()
		mmf.Menu.Disable("EditUndo")
	end
	
	if (self.undoStack:RedoAvailable()) then
		Controls.toolbar.tbRedo:Enable()
		mmf.Menu.Enable("EditRedo")
	else
		Controls.toolbar.tbRedo:Disable()
		mmf.Menu.Disable("EditRedo")
	end

	-- Step individual tools

end