local util		= require "./lua/util"

module (..., package.seeall)

DIALOG_CANCEL = 2
DIALOG_NO = 5
DIALOG_YES = 9

function ActiveFilter ()
	if (editor.vars.iofilter == "") then
		return Profile.defaultIOFilter
	end
	
	return editor.vars.iofilter
end

function SetupImportFilters ()
	local filterstrings = { }
	
	local active = ActiveFilter()
	local defid = 1
	
	for i,f in ipairs(Conf.ImportFilters) do
		filterstrings[i] = f.name .. "|" .. f.filter
		if (active == f.id) then
			defid = i
		end
	end
	
	local fstr = util.str_join("|", filterstrings)
	
	MF_SetFileFilters(fstr, defid)
end

function SetupExportFilters ()
	local filterstrings = { }
	
	local active = ActiveFilter()
	local defid = 1
	
	for i,f in ipairs(Conf.ExportFilters) do
		filterstrings[i] = f.name .. "|" .. f.filter
		if (active == f.id) then
			defid = i
		end
	end
	
	local fstr = util.str_join("|", filterstrings)
	
	MF_SetFileFilters(fstr, defid)
end

-- ----------------------------------------------------------------------------

local IOFilter = { }

function SaveAs ()
	SetupExportFilters()
	local filepath, filefilter = MF_SaveFileDialog ()
	if (filepath ~= nil and filepath ~= "") then
		if (filefilter ~= nil) then
			if (tonumber (filefilter) == 2 and string.find (filepath, "%.tlv$") == nil) then
				filepath = filepath .. ".tlv"
			end
		end

		editor.vars.filepath = filepath
		editor.vars.filename = util.basename(filepath)
		
		if (IOFilter == nil or editor.vars.iofilter ~= filefilter) then
			editor.vars.iofilter = filefilter
			IOFilter = require "./lua/io/tlv"
			IOFilter.editor = editor
		end

		local stat, err = pcall(function() return Export() end)
		if (stat == false) then
			MF_Error(err)
			return false
		end
		
		editor.ClearModified()
		return true
	end
	
	return false
end

function Save ()
	if (editor.vars.filepath == "") then
		return SaveAs()
	else
		local stat, err = pcall(function() return Export() end)
		if (stat == false) then
			MF_Error(err)
			return false
		end
		
		editor.ClearModified()
		return true
	end
end

function OpenGivenFile (filename)
	if (not PromptSaveFile() ) then
		return
	end

	--IOC:reset()
	--MF_ResetFrame ()
	
	if (IOFilter == nil or editor.vars.iofilter ~= filefilter) then
		editor.vars.iofilter = "tlv"
		IOFilter = require "./lua/io/tlv"
		IOFilter.editor = editor
	end

	local stat, err = pcall (function () return Import(filename) end)
	if (stat == false) then
		MF_Error (err)
	end
	
	editor.vars.filepath = filename
	editor.vars.filename = util.basename(filename)
	
	---DoCall ("MF_SetFileName", "Tile Lattice - " .. Editor.current_name)
	--MF_SetFileName ("Tile Lattice - " .. Editor.current_name)
	
	editor.ClearModified()
end

function Open ()
	--if (not PromptSaveFile() ) then
	--	return
	--end
	SetupImportFilters()

	local filename = MF_OpenFileDialog ()
	if (filename == nil or filename == "") then
		return
	end

	OpenGivenFile (filename)
end

function PromptSaveFile ()
	if (editor.vars.modified == true) then
		local ans = MF_SaveChangesDialog()
		if (ans == DIALOG_YES) then
			return Save()
		elseif (ans == DIALOG_NO) then
			return true
		else
			return false
		end
	end
	
	return true
end

-- ----------------------------------------------------------------------------

function Import (filename)
	local handle = io.open(filename, "rb")
	io.input(handle)
	
	local filedata = io.read("*all")
	io.close()
	
	IOFilter.DoImport(filedata)
	
	io.input(io.stdin)
end

function Export ()
	local handle = io.open(editor.vars.filename, "wb")
	io.output(handle)
	
	local filedata = IOFilter.DoExport()
	
	io.write(filedata)
	io.flush()
	io.close()
	io.output(io.stdout)
end