local util = require "./lua/util"

local string = string
local os = os

local setmetatable = setmetatable
local DoCall = DoCall

local mmf = mmf

local MF_SetFrameSize = MF_SetFrameSize
local MF_SetChipsets = MF_SetChipsets
local MF_SetFrameColor = MF_SetFrameColor

module (...)

-- ----------------------------------------------------------------------------

function FrameClass ()
	return Frame:New()
end

Frame = {
	frameWidth = 480,
	frameHeight = 480,
	tileWidth = 16,
	tileHeight = 16,
	dataDir = "./data/",
	levelName = "Untitled",
	author = "",
	key = "",
	backgroundColor = 0xFFFFFF,
	dateCreated = os.time(),
	dateModified = os.time(),
	
	updateQueued = false,
	tempdata = { },
}

function Frame:New ()
	obj = {}
	setmetatable (obj, self)
	self.__index = self
	self:Init()
	return obj
end

function Frame:Init ()
	self.frameWidth = 480
	self.frameHeight = 480
	self.tileWidth = 16
	self.tileHeight = 16
	self.levelName = "Untitled"
	self.author = ""
	self.backgroundColor = 0xFFFFFF
	self.dateCreated = os.time()
	self.dateModified = os.time()
	
	self.updateQueued = false
	self.tempdata = { }
end

function Frame:OpenProperties ()
	self.tempdata = { }
	mmf.Dialog.OpenFrameProperties()
end

function Frame:QueueUpdate ()
	self.updateQueued = true
end

function Frame:UpdateProperties ()
	self.levelName = self.tempdata.levelName
	self.author = self.tempdata.author
	self.backgroundColor = self.tempdata.backgroundColor
	
	if (self.tempdata.frameWidth > 0) then
		self.frameWidth = self.tempdata.frameWidth
	end
	if (self.tempdata.frameHeight > 0) then
		self.frameHeight = self.tempdata.frameHeight
	end
	
	if (self.tempdata.tileWidth > 0) then
		self.tileWidth = self.tempdata.tileWidth
	end
	if (self.tempdata.tileHeight > 0) then
		self.tileHeight = self.tempdata.tileHeight
	end
	
	self.updateQueued = false
end

function Frame:DateCreated ()
	return os.date("%B %d, %Y - %I:%M:%S %p", self.dateCreated)
end

function Frame:GetProperties ()
	return self.frameWidth, self.frameHeight, self.collisionChip, self.backgroundChip, self.foregroundChip, self.dataDir, self.tileWidth, self.tileHeight, self.background_color
end
	
function Frame:Import (data, ver)
	self.frameWidth = util.unu16(string.sub(data, 1, 2))
	self.frameHeight = util.unu16(string.sub(data, 3, 4))
	self.collisionChip = util.unfixstr(string.sub(data, 5, 68))
	self.backgroundChip = util.unfixstr(string.sub(data, 69, 132))
	self.foregroundChip = util.unfixstr(string.sub(data, 133, 196))
	
	if (ver >= 1) then
		self.tileWidth = util.unu16(string.sub(data, 197, 198))
		self.tileHeight = util.unu16(string.sub(data, 199, 200))
	end
	
	if (ver >= 2) then
		self.levelName = util.unfixstr(string.sub(data, 201, 264))
		self.author = util.unfixstr(string.sub(data, 265, 296))
		self.key = util.unfixstr(string.sub(data, 297, 360))
	end
	
	if (ver >= 3) then
		self.background_color = util.unu32(string.sub(data, 361, 364))
	end

	MF_SetFrameSize (self.frameWidth, self.frameHeight, self.tileWidth, self.tileHeight)
	MF_SetChipsets (self.dataDir .. self.collisionChip, self.dataDir .. self.backgroundChip, self.dataDir .. self.foregroundChip)
	MF_SetFrameColor (self.background_color)
end
	
function Frame:Export ()
	local data = self.filep.mime .. util.u16(self.filep.ver) .. util.u16(self.filep.hsz)
	data = data .. util.u16(self.frameWidth) .. util.u16(self.frameHeight)
	
	data = data .. util.fixstr( self.collisionChip, 64 )
	data = data .. util.fixstr( self.backgroundChip, 64 )
	data = data .. util.fixstr( self.foregroundChip, 64 )
	
	data = data .. util.u16(self.tileWidth) .. util.u16(self.tileHeight)
	
	data = data .. util.fixstr (self.levelName, 64)
	data = data .. util.fixstr (self.author, 32)
	data = data .. util.fixstr (self.key, 64)
	
	data = data .. util.u32 (self.background_color)
	
	return data
end
	
function Frame:Reset ()

end