local Container = require "./lua/container"
local setmetatable = setmetatable

module (...)

-- ----------------------------------------------------------------------------

UndoStack = { }

function UndoStack:New ()
	obj = {	
		ustack = Container.SlidingDeque:New(50),
		rstack = Container.SlidingDeque:New(50),
	}
	setmetatable(obj, self)
	self.__index = self
	return obj
end

function UndoStack:LogData (data)
	self.rstack:Clear()
	self.ustack:PushBack(data)
end

function UndoStack:LogUndo (data)
	self.ustack:PushBack(data)
end

function UndoStack:LogRedo (data)
	self.rstack:PushBack(data)
end

function UndoStack:Undo ()
	if (self.ustack:Empty() ) then
		return nil
	end
	return self.ustack:PopBack()
end

function UndoStack:Redo ()
	if (self.rstack:Empty() ) then
		return nil
	end
	return self.rstack:PopBack()
end

function UndoStack:UndoAvailable ()
	if (self.ustack:Empty() ) then
		return false
	end
	return true
end

function UndoStack:RedoAvailable ()
	if (self.rstack:Empty() ) then
		return false
	end
	return true
end

-- ----------------------------------------------------------------------------

UndoManager = { }

function UndoManager:New ()
	obj = {	
		undoStack = UndoStack:New(),
		dataPool = { },
		recording = false,
	}
	setmetatable(obj, self)
	self.__index = self
	return obj
end

function UndoManager:StartRecording ()
	dataPool = { }
	self.recording = true
end

function UndoManager:StopRecording ()
	self.recording = false
	self.undoStack:LogData(self.dataPool)
end

function UndoManager:LogData (k, v)
	dataPool[k] = v
end
